////////////////////////////////////////////////////////////////////////
/// \file  ChannelMapLArIATAlg.h
/// \brief Interface to algorithm class for a specific detector channel mapping
///
/// \version $Id:  $
/// \author  brebel@fnal.gov
////////////////////////////////////////////////////////////////////////
#ifndef GEO_CHANNELMAPLARIATALG_H
#define GEO_CHANNELMAPLARIATALG_H

#include <vector>
#include <set>

#include "SimpleTypesAndConstants/RawTypes.h" // raw::ChannelID_t
#include "SimpleTypesAndConstants/geo_types.h"
#include "Geometry/ChannelMapAlg.h"
#include "Geo/GeoObjectSorterLArIAT.h"
#include "fhiclcpp/ParameterSet.h"

#include "TVector3.h"

namespace geo{

  class ChannelMapLArIATAlg : public ChannelMapAlg{

  public:

    ChannelMapLArIATAlg(fhicl::ParameterSet const& p);
    
    void                     Initialize( GeometryData_t& geodata ) override;
    void                     Uninitialize();
    
    std::vector<WireID>      ChannelToWire(raw::ChannelID_t channel)           const;
    unsigned int             Nchannels()                               const;
    //@{
    virtual double WireCoordinate
      (double YPos, double ZPos, geo::PlaneID const& planeID) const override;
    virtual double WireCoordinate(double YPos, double ZPos,
                                 unsigned int PlaneNo,
                                 unsigned int TPCNo,
                                 unsigned int cstat) const
      { return WireCoordinate(YPos, ZPos, geo::PlaneID(cstat, TPCNo, PlaneNo)); }
    //@}
    
    //@{
    virtual WireID NearestWireID
      (const TVector3& worldPos, geo::PlaneID const& planeID) const override;
    virtual WireID NearestWireID(const TVector3& worldPos,
                                 unsigned int    PlaneNo,
                                 unsigned int    TPCNo,
                                 unsigned int    cstat) const override
      { return NearestWireID(worldPos, geo::PlaneID(cstat, TPCNo, PlaneNo)); }
    //@}
    
    //@{
    virtual raw::ChannelID_t PlaneWireToChannel
      (geo::WireID const& wireID) const override;
    virtual raw::ChannelID_t PlaneWireToChannel(unsigned int plane,
                                                unsigned int wire,
                                                unsigned int tpc,
                                                unsigned int cstat) const override
      { return PlaneWireToChannel(geo::WireID(cstat, tpc, plane, wire)); }
    //@}

   View_t                    View( raw::ChannelID_t const channel )            const;
   SigType_t                 SignalType( raw::ChannelID_t const channel )      const;
   std::set<View_t>  const&  Views()                                   const;
   std::set<PlaneID> const&  PlaneIDs()                                const;

   // methods for the auxiliary detectors
/*
   // method returns the entry in the sorted AuxDetGeo vector so that the 
   // Geometry in turn can return that object
   virtual size_t            NearestAuxDet         (TVector3 const& point)   const override;
   virtual size_t            NearestSensitiveAuxDet(TVector3 const& point)   const override;
   virtual size_t            NSensitiveAuxDet(std::string const& auxDetName) const override;
*/
  private:
    
    unsigned int                                         fNcryostat;      ///< number of cryostats in the detector
    unsigned int                                         fNchannels;      ///< number of channels in the detector
    raw::ChannelID_t                                     fTopChannel;     ///< book keeping highest channel #
    std::vector<unsigned int>                            fNTPC;           ///< number of TPCs in each cryostat
    std::set<View_t>                                     fViews;          ///< vector of the views present in the detector
    std::set<PlaneID>                                    fPlaneIDs;       ///< vector of the PlaneIDs present in the detector
    std::vector<std::vector<std::vector<float>>>         fFirstWireProj;  ///< Distance (0,0,0) to first wire 	 
                                                                          ///< along orth vector per plane per TPC
    std::vector<std::vector<std::vector<float>>>         fOrthVectorsY;   ///< Unit vectors orthogonal to wires in
    std::vector<std::vector<std::vector<float>>>         fOrthVectorsZ;   ///< each plane - stored as 2 components
                                                                          ///< to avoid having to invoke any bulky
                                                                          ///< TObjects / CLHEP vectors etc	 
    std::vector<std::vector<std::vector<float>>>         fWireCounts;     ///< Number of wires in each plane - for
                                                                          ///< range checking after calculation   
    std::vector<std::vector<unsigned int>>  		 fNPlanes;        ///< Number of planes in each TPC - for
                                                                          ///< range checking after calculation   
    std::vector<std::vector<std::vector<unsigned int>>>  fPlaneBaselines; ///< The number of wires in all the 
                                                                          ///< tpcs and planes up to this one 
                                                                          ///< in the heirachy
    std::vector<std::vector<std::vector<unsigned int>>>  fWiresPerPlane;  ///< The number of wires in this plane 
                                                                          ///< in the heirachy
    std::vector<std::string>                             fAuxDetNames;    ///< names of the sorted auxiliary detectors
    std::vector<std::vector<float>>                      fAuxDetCenters;  ///< centers of the sorted auxiliary detectors
                                                                          ///< in world coordinates
    std::vector<std::vector<float>>                      fAuxDetHalfExt;  ///< half width, height, and length of the sorted
                                                                          ///< auxiliary detectors in world coordinates
    std::vector<std::vector<uint32_t>>                   fAuxDetSensitive;///< number of sensitive volumes in each auxiliary detector
    std::vector<std::vector<std::vector<float>>>         fADSensCenters;  ///< centers of the channels in the sorted auxiliary 
                                                                          ///< detector in world coordinates
    std::vector<std::vector<std::vector<float>>>         fADSensHalfExt;  ///< half width, height, and length of the channels 
                                                                          ///< in the sorted auxiliary detectors in world coordinates
    geo::GeoObjectSorterLArIAT                           fSorter;         ///< class to sort geo objects
  };


}
#endif // GEO_CHANNELMAPLARIATALG_H

